%{
#include "c.h"

#define WORDSIZE 2
#define LONGSIZE 4

#define NODEPTR_TYPE Node
#define OP_LABEL(p) ((p)->op)
#define LEFT_CHILD(p) ((p)->kids[0])
#define RIGHT_CHILD(p) ((p)->kids[1])
#define STATE_LABEL(p) ((p)->x.state)

#define isdouble(t) (isfloat(t) && (t)->size == 8)

#define blop(n) (want_tmpregs = 1, (n))
    
    static int cstate;
    static int argoffset;
    
    static void _trace(NODEPTR_TYPE, int, int, int);

    static void stopconst(void);
    static char* mystringd(int);

    static void address(Symbol, Symbol, long);
    static void blkfetch(int, int, int, int);
    static void blkloop(int, int, int, int, int, int[]);
    static void blkstore(int, int, int, int);
    static void defaddress(Symbol);
    static void defconst(int, int, Value);
    static void defstring(int, char *);
    static void defsymbol(Symbol);
    static void doarg(Node);
    static void emit2(Node);
    static void export(Symbol);
    static void progend(void);
    static void progbeg(int, char **);
    static void clobber(Node);
    static void function(Symbol, Symbol [], Symbol [], int);
    static void global(Symbol);
    static void import(Symbol);
    static void local(Symbol);
    static void segment(int);
    static void space(int);
    static void target(Node);

    /* Floating point garbage: */
    static void build_glulx_float(double f, unsigned int *w1);
    static void build_glulx_double(double f, unsigned int *w1, 
				   unsigned int *w2);    

    static Symbol rreg[32], rreg2[32];
    static int tmpregs[] = {0, 1, 2};
    static Symbol rregw, rreg2w;
    static Symbol spreg;

    static int cseg, retstruct, want_tmpregs;
    
    /* 
     * Notes:
     * 
     * F suffixed operators aren't implemented.
     * Unsigned relative operators aren't correct!
     * Unsigned division and modulus operators aren't correct!
     */
%}
%start stmt

%term CNSTI1=1045
%term CNSTU1=1046
%term CNSTI2=2069
%term CNSTU2=2070
%term CNSTP2=2071
%term CNSTI4=4117
%term CNSTU4=4118
%term CNSTF4=4113
%term CNSTF8=8209
 
%term ARGB=41
%term ARGI2=2085
%term ARGU2=2086
%term ARGP2=2087
%term ARGF4=4129
%term ARGF8=8225

%term ASGNB=57
%term ASGNI1=1077
%term ASGNU1=1078
%term ASGNI2=2101
%term ASGNU2=2102
%term ASGNP2=2103
%term ASGNI4=4149
%term ASGNU4=4150
%term ASGNF4=4145
%term ASGNF8=8241

%term INDIRB=73
%term INDIRI1=1093
%term INDIRU1=1094
%term INDIRI2=2117
%term INDIRU2=2118
%term INDIRP2=2119
%term INDIRI4=4165
%term INDIRU4=4166
%term INDIRF4=4161
%term INDIRF8=8257

%term CVFF4=4209
%term CVFI4=4213
%term CVFF8=8305

%term CVII1=1157
%term CVIU1=1158
%term CVII2=2181
%term CVIU2=2182
%term CVII4=4229
%term CVIU4=4230
%term CVIF4=4225
%term CVIF8=8321

%term CVPU2=2198

%term CVUI1=1205
%term CVUU1=1206
%term CVUI2=2229
%term CVUU2=2230
%term CVUP2=2231
%term CVUI4=4277
%term CVUU4=4278

%term NEGI2=2245
%term NEGF4=4289
%term NEGF8=8385

%term CALLV=216
%term CALLB=217
%term CALLI2=2261
%term CALLU2=2262
%term CALLP2=2263
%term CALLF4=4305
%term CALLF8=8401

%term RETV=248
%term RETI2=2293
%term RETU2=2294
%term RETP2=2295
%term RETF4=4337
%term RETF8=8433

%term ADDRGP2=2311
%term ADDRFP2=2327
%term ADDRLP2=2343

%term ADDI2=2357
%term ADDU2=2358
%term ADDP2=2359
%term ADDF4=4401
%term ADDF8=8497

%term SUBI2=2373
%term SUBU2=2374
%term SUBP2=2375
%term SUBF4=4417
%term SUBF8=8513

%term LSHI2=2389
%term LSHU2=2390
%term MODI2=2405
%term MODU2=2406
%term RSHI2=2421
%term RSHU2=2422
%term BANDI2=2437
%term BANDU2=2348
%term BCOMI2=2453
%term BCOMU2=2454
%term BORI2=2469
%term BORU2=2470
%term BXORI2=2485
%term BXORU2=2486

%term DIVI2=2501
%term DIVU2=2502
%term DIVF4=4545
%term DIVF8=8641

%term MULI2=2517
%term MULU2=2518
%term MULF4=4561
%term MULF8=8657

%term EQF4=4577
%term EQF8=8673
%term EQI2=2533
%term EQU2=2534

%term GEF4=4593
%term GEF8=8689
%term GEI2=2549
%term GEU2=2548

%term GTF4=4609
%term GTF8=8705
%term GTI2=2565
%term GTU2=2566

%term LEF4=4625
%term LEF8=8721
%term LEI2=2581
%term LEU2=2582

%term LTF4=4641
%term LTF8=8737
%term LTI2=2597
%term LTU2=2598

%term NEF4=4657
%term NEF8=8753
%term NEI2=2613
%term NEU2=2614

%term JUMPV=584

%term LABELV=600

%term LOADB=233
%term LOADF4=4321
%term LOADF8=8417
%term LOADI1=1253
%term LOADI2=2277
%term LOADI4=4325
%term LOADP2=2279
%term LOADU1=1254
%term LOADU2=2278
%term LOADU4=4326

%term VREGP=711
%%
reg:  INDIRI1(VREGP)		"# read register\n"
reg:  INDIRU1(VREGP)		"# read register\n"

reg:  INDIRI2(VREGP)		"# read register\n"
reg:  INDIRU2(VREGP)		"# read register\n"

reg:  INDIRF4(VREGP)		"# read register\n"
reg:  INDIRI4(VREGP)		"# read register\n"
reg:  INDIRP2(VREGP)		"# read register\n"
reg:  INDIRU4(VREGP)		"# read register\n"

reg:  INDIRF8(VREGP)		"# read register\n"

stmt: ASGNI1(VREGP,reg)		"# write register\n"
stmt: ASGNU1(VREGP,reg)		"# write register\n"
stmt: ASGNI2(VREGP,reg)		"# write register\n"
stmt: ASGNU2(VREGP,reg)		"# write register\n"
stmt: ASGNP2(VREGP,reg)		"# write register\n"
stmt: ASGNF4(VREGP,reg)		"# write register\n"
stmt: ASGNI4(VREGP,reg)		"# write register\n"
stmt: ASGNU4(VREGP,reg)		"# write register\n"
stmt: ASGNF8(VREGP,reg)		"# write register\n"

icon: CNSTI1			"%a"
icon: CNSTU1			"%a"
icon: CNSTI2			"%a"
icon: CNSTU2			"%a"
icon: CNSTP2			"%a"
icon: CNSTI4			"%a"
icon: CNSTU4			"%a"

con:  icon			"%0"
con:  ADDRGP2			"%a"
stmt: reg			""

reg:  ADDRGP2			"\tr%c = %a;\n"			1
reg:  ADDRLP2			"\tr%c = xp + %a;\n"	       		1
reg:  ADDRFP2			"\tr%c = xp + %a;\n"	       		1

reg:  INDIRI1(rc)		"\t@loadb %0 0 -> r%c;\n"		1
reg:  INDIRU1(rc)		"\t@loadb %0 0 -> r%c;\n"		1
reg:  INDIRI2(rc)		"\t@loadw %0 0 -> r%c;\n" 		1
reg:  INDIRU2(rc)		"\t@loadw %0 0 -> r%c;\n" 		1
reg:  INDIRI4(rc)		"load %0, r%c\n"			1
reg:  INDIRU4(rc)		"load %0, r%c\n"			1
reg:  INDIRP2(rc)		"\t@loadw %0 0 -> r%c;\n" 		1
reg:  INDIRF4(rc)		"load %0, r%c\n"			1

reg:  INDIRI1(ADDP2(rc, icon))	"\t@loadb %0 %1 -> r%c;\n"		1
reg:  INDIRU1(ADDP2(rc, icon))	"\t@loadb %0 %1 -> r%c;\n"    		1
reg:  INDIRI2(ADDP2(rc, icon))	"\t@loadw %0 %1/2 -> r%c;\n" 		1
reg:  INDIRU2(ADDP2(rc, icon))	"\t@loadw %0 %1/2 -> >%c;\n"		1
reg:  INDIRP2(ADDP2(rc, icon))	"\t@loadw %0 %1/2 -> r%c;\n"		1
reg:  INDIRI4(ADDP2(rc, icon))	"\tload %0, %1/4, r%c\n"     		1
reg:  INDIRU4(ADDP2(rc, icon))	"\tload %0, %1/4, r%c\n"     		1
reg:  INDIRF4(ADDP2(rc, icon))	"\tload %0, %1/4, r%c\n"		1

reg:  ASGNI1(rcg, rcg)		"\t@storeb %0 0 %1;\n"			1
reg:  ASGNI2(rcg, rcg)		"\t@storew %0 0 %1;\n" 			1
reg:  ASGNI4(rcg, rcg)		"\tstore %0 %1;\n"			1
reg:  ASGNU1(rcg, rcg)		"\t@storeb %0 0 %1;\n"			1
reg:  ASGNU2(rcg, rcg)		"\t@storew %0 0 %1;\n"			1
reg:  ASGNU4(rcg, rcg)		"\tstore %0 %1;\n"			1
reg:  ASGNP2(rcg, rcg)		"\t@storew %0 0 %1;\n"			1
reg:  ASGNF4(rcg, rcg)		"\tstore %0 %1;\n"			1

reg:  INDIRF8(rcg)		"# INDIRF8\n"				2
stmt: ASGNF8(rcg, rcg)		"# ASGNF8\n"				2

rc:   con			"%0"
rc:   reg			"r%0"

rcg:  rc			"%0"

reg:  con			"\tr%c = %0;\n"			1

reg:  ADDI2(rcg,rcg)		"\tr%c = %0 + %1;\n"			1
reg:  ADDU2(rcg,rcg)		"\tr%c = %0 + %1;\n"			1
reg:  ADDP2(rcg,rcg)		"\tr%c = %0 + %1;\n"			1
reg:  BANDI2(rcg,rcg)		"\t@and %0 %1 -> r%c;\n"       		1
reg:  BORI2(rcg,rcg)		"\t@or %0 %1 -> r%c;\n"			1
reg:  BXORI2(rcg,rcg)		"\t@xor %0 %1 -> r%c;\n"       		1
reg:  BANDU2(rcg,rcg)		"\t@and %0 %1 -> r%c;\n"       		1
reg:  BORU2(rcg,rcg)		"\t@or %0 %1 -> r%c;\n"			1
reg:  BXORU2(rcg,rcg)		"\t@xor %0 %1 -> r%c;\n"		1
reg:  SUBI2(rcg,rcg)		"\t@sub %0 %1 -> r%c;\n"		1
reg:  SUBP2(rcg,rcg)		"\t@sub %0 %1 -> r%c;\n"		1
reg:  SUBU2(rcg,rcg)		"\t@sub %0 %1 -> r%c;\n"		1
reg:  LSHI2(rcg,rcg)		"\t@art_shift %0 %1 -> r%c;\n" 		1
reg:  LSHU2(rcg,rcg)		"\t@log_shift %0 %1 -> r%c;\n" 		1
reg:  RSHI2(rcg,rcg)		"\t@art_shift %0 0-%1 -> r%c;\n"	1
reg:  RSHU2(rcg,rcg)		"\t@log_shift %0 0-%1 -> r%c;\n"    	1
reg:  NEGI2(rcg)		"\t@mul %0 0-1 -> r%c;\n"		1
reg:  BCOMI2(rcg)		"\t@not %0 -> r%c;\n"			1
reg:  BCOMU2(rcg)		"\t@not %0 -> r%c;\n"			1
reg:  DIVI2(rcg,rcg)		"\t@div %0 %1 -> r%c;\n"		1
reg:  DIVU2(rcg,rcg)		"\t@div %0 %1 -> r%c;\n"		1
reg:  MODI2(rcg,rcg)		"\t@mod %0 %1 -> r%c;\n"		1
reg:  MODU2(rcg,rcg)		"\t@mod %0 %1 -> r%c;\n"		1
reg:  MULI2(rcg,rcg)		"\t@mul %0 %1 -> r%c;\n"		1
reg:  MULU2(rcg,rcg)		"\t@mul %0 %1 -> r%c;\n"		1

reg:  CVII4(reg)       		"# extend\n"				2
reg:  CVUU4(reg)		"# extend\n"				2
reg:  CVUI4(reg)		"# extend\n"				2
reg:  CVIU4(reg)		"# extend\n"				2
reg:  CVIU2(reg)		"# truncate\n"				2
reg:  CVII2(reg) 		"# truncate\n" 				2
reg:  CVUU2(reg)		"# truncate\n"				2
reg:  CVUI2(reg)		"# truncate\n"				2
reg:  CVPU2(reg)		"# truncate\n"				2
reg:  CVIU1(reg)		"# truncate\n"				2
reg:  CVII1(reg)       		"# truncate\n"				2
reg:  CVUU1(reg)		"# truncate\n"				2
reg:  CVUI1(reg)		"# truncate\n"				2

reg:  CVFF4(reg)		"# CVFF4\n"				5
reg:  CVFF8(reg)		"# CVFF8\n"				5
reg:  CVFI4(reg)		"# CVFI4\n"				5
reg:  CVIF4(reg)		"# CVIF4\n"				5
reg:  CVIF8(reg)		"# CVIF8\n"				5

reg:  LOADI1(rcg)  		"\tr%c = %0 & $FF;\n"		move(a)
reg:  LOADI2(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADI4(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADP2(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADU1(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADU2(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADU4(rcg)  		"\tr%c = %0;\n"  		move(a)
reg:  LOADF4(rcg)  		"\t7@store r%c %0;\n"  		move(a)

stmt: JUMPV(rcg)		"\tjump %0;\n"
stmt: LABELV			".%a;\n"

stmt: EQI2(rcg,rc)		"\t@je %0 %1 ?%a;\n"			1
stmt: EQU2(rcg,rc)		"\t@je %0 %1 ?%a;\n"			1
stmt: GEI2(rcg,rc)		"\t@jl %0 %1 ?~%a;\n"			1
stmt: GEU2(rcg,rc)		"jge %0, %1, %a\n"			1
stmt: GTI2(rcg,rc)		"\t@jg %0 %1 ?%a;\n"			1
stmt: GTU2(rcg,rc)		"jgt %0, %1, %a\n"			1
stmt: LEI2(rcg,rc)		"\t@jg %0 %1 ?~%a;\n"			1
stmt: LEU2(rcg,rc)		"jle %0, %1, %a\n"			1
stmt: LTI2(rcg,rc)		"\t@jl %0 %1 ?%a;\n"			1
stmt: LTU2(rcg,rc)		"jlt %0, %1, %a\n"			1
stmt: NEI2(rcg,rc)		"\t@je %0 %1 ?~%a;\n"			1
stmt: NEU2(rcg,rc)		"\t@je %0 %1 ?~%a;\n"			1

stmt: CALLV(rc)			"# CALLV\n"				2
stmt: CALLB(rc,reg)		"# CALLB\n"				blop(3)
reg:  CALLI2(rc)		"# CALLI4\n"				2
reg:  CALLP2(rc)		"# CALLP4\n"				2
reg:  CALLU2(rc)		"# CALLU4\n"				2
reg:  CALLF4(rc)		"# CALLF4\n"				2
reg:  CALLF8(rc)		"# CALLF8\n"				3

stmt: RETI2(reg)       		"\t@ret r%0;\n"				1
stmt: RETU2(reg)       		"\t@ret r%0;\n"				1
stmt: RETP2(reg)       		"\t@ret r%0;\n"				1
stmt: RETF4(reg)       		"# ret\n"				1
stmt: RETF8(reg)		"# ret\n"				1

stmt: ARGI2(reg)		"# ARGI2\n"				1
stmt: ARGU2(reg)		"# ARGU2\n"				1
stmt: ARGP2(reg)		"# ARGP2\n"				1
stmt: ARGF4(reg)		"# ARGF4\n"				1
stmt: ARGF8(reg)		"# ARGF8\n"				1

stmt: ASGNB(reg, INDIRB(reg))	"# ASGNB\n"				blop(0)

%%

static char* mystringd(int v)
{
	return stringf("0%s%D", (v>=0) ? "+" : "", v);
}

static void progend(void)
{
	stopconst();
}

static void progbeg(int argc, char *argv[]) 
{
    int i;
    
    /* Yucky.  Determine host endianess: */
    {
	union {
	    char c;
	    int i;
	} u;
	u.i = 0;
	u.c = 1;
	swap = ((int)(u.i == 1)) != IR->little_endian;
    }
    
    parseflags(argc, argv);
    
    /* We don't *really* have 32 registers, but I'd rather just use the
     * facilities provided by the (fairly ill-fitting) general generator
     * code. */
    for (i=0; i<15; i++)
      rreg[i] = mkreg("%d", i, 1, IREG);

#if 0
    for (i=0; i<15; i++)
      rreg2[i] = mkreg("%d", i, 3, IREG);
#endif

    rregw = mkwildcard(rreg);
//    rreg2w = mkwildcard(rreg2);
    
    spreg = mkreg("sp", 1, 1, FREG);
    
    tmask[0] = 0xffffffff; tmask[1] = 0;
    vmask[0] = 0x00ffffff; vmask[1] = 0;

    /* Not emitting a constant. */

    cstate = 0;
}

static Symbol rmap(int opk) 
{
	switch (optype(opk))
	{
		case I:
		case U:
		case P:
		case B:
			return rregw;

#if 0
    case F:      
      return (opsize(opk) == 8) ? rreg2w : rregw;
#endif
	}
	return 0;
}

static void target(Node p) 
{
	/* Not much targetting to do. */
}

static void clobber(Node p) 
{
	/* Not much clobbering to do. */
}

static void doarg(Node p) 
{
    /* Unfortunately lcc doesn't provide a hook for docall().
     * I abuse mkactual() to record arguments.  This modifies
     * argoffset, which gets stashed in syms[0] of a call node
     * when the internal docall() is reached. */
    assert(p->op != ARG + B);
    mkactual(1, p->op == ARG+F+sizeop(8) ? 2 : 1);    
}

static void emitrc(Node p)
{
  switch (generic(p->op)) {
  case CNST:
    print("%s", p->syms[0]->x.name);
    break;
  case ADDRG:
    print("%s", p->syms[0]->x.name);
    break;
  default:
    assert(p->x.registered);
    print("%%r%d", getregnum(p));
  }
}

static void emitcall(Node p, int argc, int reg)
{
	print("\t@sub ");
  	if (framesize)
		print("nxp ");
	else
		print("xp ");
	print("%d sp;\n", argoffset);
	
	print("\t@call_v");
	if (reg != -1)
		print("s ");
	else
		print("n ");

	emitrc(p);
	print(" sp");
  
	if (reg != -1)
		print(" -> r%d", reg);
	print(";\n");

	argoffset = 0;
}

static void emit2(Node p)
{
  /* FIXME: this relies on strict right-to-left evaluation of function
   * arguments.  Should check that this is actually being followed. */

  /* FIXME: why can't I use p->x.kids here? */

  static char sizechar[] = "?bh? ";

  int p_opsize = opsize(p->op), size, preg, kreg;

  switch(specific(p->op)) {
  case RET+F:
    if (opsize(p->op) == 8) {
      assert(0);
    }
    break;

		case ARG+I:
		case ARG+U:
		case ARG+P:
		case ARG+F:
		{
			char* sp;
			int reg = getregnum(p->kids[0]);

			if (framesize)
				sp = "nxp";
			else
				sp = "xp";

			print("\t! push arg r%d\n", reg);
			switch(opsize(p->op))
			{
				case 1:
					argoffset += 1;
					print("\t@store %s 0-%d r%d;\n",
						sp, argoffset, reg);
					break;

				case 2:
					argoffset += 2;
					print("\t@storew %s 0-%d/2 r%d;\n",
						sp, argoffset, reg);
					break;

				case 4:
					print("\tunimplemented ARG\n");
					break;
			}
		}
		break;

#if 0
	case ARG+F:
  	case ARG+I:
	case ARG+P:
	case ARG+U:
	{
		break;
		print("! push arg r%d\n", getregnum(p->kids[0]));
		break;
	}
      if (opsize(p->op) == 8) {
	  print("push %%r%d\n", getregnum(p->kids[0] + 1));
	  print("push %%r%d\n", getregnum(p->kids[0]));
      }
      break;
#endif
  case ASGN+B:
    dalign = salign = p->syms[1]->u.c.v.i;
    blkcopy(getregnum(p->x.kids[0]), 0,
	    getregnum(p->x.kids[1]), 0,
	    p->syms[0]->u.c.v.i, tmpregs);
    break;
  case CALL+F:
    if (opsize(p->op) == 8) {
      assert(0);
      break;
    }
    /* Fall through */
  case CALL+I: case CALL+U: case CALL+P:
    emitcall(p->kids[0], p->syms[0]->u.c.v.i, getregnum(p));
    break;
  case CALL+B:
    print("push %%r%d\n", getregnum(p->kids[1]));
    emitcall(p->kids[0], p->syms[0]->u.c.v.i, getregnum(p));
    break;
  case CALL+V:
    emitcall(p->kids[0], p->syms[0]->u.c.v.i, -1);
    break;

		case CVU+I: 
		case CVI+I:
		case CVP+I:
		{
			int in = getregnum(p->x.kids[0]);
			int out = getregnum(p);
			int insize = p->syms[0]->u.c.v.i;

			print("\t! src=%d dest=%d\n", in, out);
			if ((p_opsize == 2) && (insize == 1))
			{
				print("\t@log_shift r%d 8 -> r%d;\n",
					in, out);
				print("\t@log_shift r%d 0-8 -> r%d;\n",
					out, out);
			}
			else if ((p_opsize == 4) && (insize == 2))
			{
				/* Hack */
				if (out != in)
					print("\tr%d = r%d;\n", out, in);
			}
			else if (p_opsize == insize)
			{
				if (in != out)
					print("\t@store r%d r%d\n",
						out, in);
			}
			else
			{
				print("\tunimplemented CV?I %d -> %d\n",
					p_opsize, insize);
			}
		}
		break;

  case CVU+U:
  case CVI+U:
  case CVP+U:
    preg = getregnum(p);
    kreg = getregnum(p->x.kids[0]);
    size = p->syms[0]->u.c.v.i;
    if (size > p_opsize && p_opsize < WORDSIZE)
      print("and %%r%d, 0x%x, %%r%d\n", kreg, 1 << (p_opsize * 8), preg);
    else if (kreg != preg)
      print("copy%c %%r%d, %%r%d\n", sizechar[size], kreg, preg);
    break;
  case INDIR+F:
    if (opsize(p->op) == 8) {
      assert(0);
    }
    break;
  case ASGN+F:
    if (opsize(p->op) == 8) {
      assert(0);
    }
    break;
  }
}

static void local(Symbol p)
{
  if (retstruct) {
    assert(p == retv);
    p->sclass = REGISTER;
    askregvar(p, rreg[0]);
    retstruct = 0;
    return;
  }

  if (isscalar(p->type) && !p->addressed && !isdouble(p->type))
    p->sclass = REGISTER;

  if (askregvar(p, rmap(ttob(p->type))) == 0)
  {
    mkauto(p);
    if (p->x.name[0] == '-')
    	p->x.name = stringf("0%s", p->x.name);
  }
}

static void function(Symbol f, Symbol caller[], Symbol callee[], int ncalls)
{
    int retstructp, i, varargs, reg;
    stopconst();
    
    retstructp = retstruct = isstruct(freturn(f->type));
    varargs = variadic(f->type);
    
    assert(!varargs);

    offset = maxoffset = maxargoffset = 0;
    usedmask[0] = usedmask[1] = 0;
    freemask[0] = freemask[1] = ~(unsigned int)0;
    want_tmpregs = 0;
    
    reg = retstruct ? 1 : 0;
    
    for (i = 0; callee[i]; i++) {
	Symbol p = callee[i], q = caller[i];
	int size = roundup(q->type->size, WORDSIZE);
	if (reg > 16) {
	    p->x.offset = q->x.offset = offset;
	    p->x.name = q->x.name = mystringd(p->x.offset);
	    p->sclass = q->sclass = AUTO;
	    print("! param %s on stack @ %s\n", p->x.name, -offset);
	} else if (p->addressed) {
	    p->x.offset = offset;
	    offset += size;
	    p->x.name = mystringd(p->x.offset);
	    p->sclass = q->sclass = AUTO;
	} else {
	    p->sclass = q->sclass = REGISTER;
	    askregvar(p, rreg[reg]);
	    q->x.name = p->x.name;
	    q->x.offset = p->x.offset = offset;
	    offset += size;
	    usedmask[0] |= 1<<reg;
	    print("! param %s cached in reg %d, on stack @ %d\n", p->name,
	    	reg, p->x.offset);
	}
	reg += isstruct(p->type) ? 1 : size / 2;
    }
    
    gencode(caller, callee);

    framesize = roundup(maxoffset, WORDSIZE);
    
    print("! .function %s\n", varargs ? "stack" : "local");
    print("[ %s", f->x.name);
    
    print(" xp  ");
    
    i = 0;
    if (retstruct) {
	assert(0);
    }
    
    /* Actual arguments */
    for (; callee[i]; i++) {
	Symbol p = callee[i];
	int size = isstruct(p->type) ? WORDSIZE : p->type->size;
    }
    
    /* Reserve space for registers; almost the inverse of saving them
     * before function entry. =)  Only those that are used are created,
     * except that it's easiest to assume that we always need r0. */
    for (i = 0; i < 32; i++) {
	if (usedmask[0] & (1 << i))
	    print(" r%d", i);
    }
    
    if (ncalls && framesize)
	print(" nxp");
    
    if (want_tmpregs)
	print(" t0 t1 t2");

    print(";\n");
    
    if (ncalls && framesize) {
	print("\t@sub xp %d nxp;\n", framesize);
    }
    
    for(i=0; caller[i]; i++)
    	if (caller[i]->sclass == REGISTER)
	{
		print("\t@loadw xp %s/2 -> r%s;\n",
			mystringd(caller[i]->x.offset), caller[i]->x.name);
	}

    emitcode();
    
    print("\t@rfalse;\n");  
    print("];\n");
}
	
static void defconst(int suffix, int size, Value v) 
{
    if (suffix == F) {
	unsigned int w1, w2;
	if (size == WORDSIZE) {
	    build_glulx_float(v.d, &w1);
	    print(".word 0x%x\n", w1);
	} else if (size == LONGSIZE) {
	    build_glulx_double(v.d, &w1, &w2);
	    print(".word 0x%x, 0x%x\n", w1, w2);	    
	} else
	    assert(0);	
    } else {
	if (suffix == P)
	{
		print("%d %d\n", ((unsigned int)v.p)>>8,
			((unsigned int)v.p) & 0xFF);
	}
	else if (size == 1)
		print("%d\n", (suffix == I) ? v.i : v.u);
	else if (size == 2)
	{
		unsigned int i = (suffix == I) ? v.i : v.u;
		print("%d %d\n", i>>8, i & 0xFF);
	}
	else assert(0);
    }	    
}

static void defaddress(Symbol p) 
{
    print(".word %s\n", p->x.name);
}

static void defstring(int n, char *str) 
{
	int i;

	for(i=0; i<n; i++)
		print(" %d", (str[i] & 0377));
	print("\n");
}

static void address(Symbol q, Symbol p, long n) 
{
    if (p->scope == GLOBAL || p->sclass == STATIC || p->sclass == EXTERN)
	q->x.name = stringf("%s%s%D", p->x.name, n >= 0 ? "+" : "", n);
    else {
	q->x.offset = p->x.offset + n;
	q->x.name = mystringd(q->x.offset);
    }    
}

static void export(Symbol p) 
{
    print("! .global %s\n", p->x.name);
}

static void import(Symbol p) 
{
    /* No action required. */
}

static void defsymbol(Symbol p) 
{
  if (p->scope >= LOCAL && p->sclass == STATIC)
    p->x.name = stringf("STATIC(%d)", genlabel(1));
  else if (p->scope >= LABELS)
    p->x.name = stringf(p->generated ? "L%s" : "_%s", p->name);
  else
    p->x.name = p->name;
}

static void segment(int n) 
{
    const char *segs = NULL;
    
    if (n == cseg)
	return;
    
    cseg = n;
    if (cseg == CODE)
	segs = "text";
    else if (cseg == BSS)
	segs = "bss";
    else if (cseg == DATA || cseg == LIT)
	segs = "data";
    print("! .seg \"%s\"\n", segs);
}

static void space(int n) 
{
	int i;
	for(i=0; i<n; i++)
		print(" 0");
	print("\n");
}

static void stopconst(void)
{
	if (!cstate)
		return;
	
	print(";\n");
	cstate = 0;
}

static void global(Symbol p)
{
	int i;

	if (p->u.seg == BSS) {
#if 0
	if (p->sclass == STATIC)
	    print(".reserve %s, %d\n", p->x.name, p->type->size);
	else 
	    print(".common %s, %d\n", p->x.name, p->type->size);
#endif
		assert(cseg == BSS);
		stopconst();
		print("Array %s ->\n", p->x.name);
		cstate = 1;
//		for(i=0; i<p->type->size; i++)
//			print(" 0");
	}
	else
	{
		stopconst();
		print("Array %s ->\n", p->x.name);
		cstate = 1;
	}
}

static void blkfetch(int k, int off, int reg, int tmp) 
{
  switch (k) {
  case 1:
    print("\t@loadb r%d %d -> t%d;\n", reg, off, tmp);
    break;
  case WORDSIZE:
    assert(off % WORDSIZE == 0);
    print("\t@loadw r%d, %d, t%d;\n", reg, off / WORDSIZE, tmp);
  case LONGSIZE:
    assert(off % LONGSIZE == 0);
    print("aload %%r%d, %d, %%t%d\n", reg, off / LONGSIZE, tmp);
    break;
  default:
    assert(k == 1 || k == WORDSIZE || k == LONGSIZE);
  }
}

static void blkstore(int k, int off, int reg, int tmp)
{
  switch (k) {
  case 1:
    print("\t@storeb r%d %d t%d;\n", reg, off, tmp);
    break;
  case WORDSIZE:
    assert(off % WORDSIZE == 0);
    print("\t@storew r%d %d t%d;\n", reg, off / WORDSIZE, tmp);
  case LONGSIZE:
    assert(off % LONGSIZE == 0);
    print("astore %%r%d, %d, %%t%d\n", reg, off / LONGSIZE, tmp);
    break;
  default:
    assert(k == 1 || k == WORDSIZE || k == LONGSIZE);
  }
}

static void blkloop(int dreg, int doff, int sreg, int soff, int size, 
		    int tmps[])
{
	char* src;
	char* dest;

	if (doff)
	{
		print("\t@add r%d %d sp;\n", dreg, doff);
		dest = "sp";
	}
		dest = stringf("r%d", dreg);

	if (soff)
	{
		print("\t@add r%d %d sp;\n", sreg, soff);
		src = "sp";
	}
	else
		src = stringf("r%d", sreg);
	
	print("\t@copy_table %s %s %d;\n", src, dest, size);

  /* This may run into problems with signed comparisons in jlt! */
  /* FIXME? */

#if 0
  print("add %%r%d, %d, %%r%d\n", sreg, size & ~7, sreg);
  print("add %%r%d, %d, %%r%d\n", dreg, size & ~7, sreg);
  blkcopy(tmps[2], doff, sreg, soff, size & 7, tmps);
  print("L%d:\n", label);
  print("sub %%r%d, 8, %%r%d\n", sreg, sreg);
  print("sub %%t%d, 8, %%t%d\n", tmps[2], tmps[2]);
  blkcopy(tmps[2], doff, sreg, soff, 8, tmps);
  print("jlt %%r%d, %%t%d\n, L%d\n", dreg, tmps[2], label);
#endif
}
	    
static void _trace(NODEPTR_TYPE p, int a, int b, int cost)
{
	print("! %s \"%s\" (b=%d cost=%d)\n", _string[a], _templates[a],
		(short)b, (short)cost);
}

Interface zcodeIR = {
    1, 1, 0,		/* char */
    WORDSIZE, 1, 0,	/* short */
    WORDSIZE, 1, 0,	/* int */
    LONGSIZE, 1, 0,	/* long */
    LONGSIZE, 1, 0,	/* long long */
    LONGSIZE, 1, 1,	/* float */
    8, 1, 1,		/* double */
    8, 1, 1,		/* long double */
    2, 1, 0,		/* pointer */
    0, 1, 0,		/* struct */
    0,			/* little_endian */
    0,			/* mulops_calls */
    1,			/* wants_callb */
    0,			/* wants_argb */
    0,			/* left_to_right */
    0,			/* wants_dag */
    0,			/* unsigned_char */
    address,
    blockbeg,
    blockend,
    defaddress,
    defconst,
    defstring,
    defsymbol,
    emit,
    export,
    function,
    gen,
    global,
    import,
    local,
    progbeg,
    progend,
    segment,
    space,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    {
	1,
	rmap,
	blkfetch, blkstore, blkloop,
	_label,
	_rule,
	_nts,
	_kids,
	_string,
	_templates,
	_isinstruction,
	_ntname,
	emit2,
	doarg,
	target,
	clobber
    }	
};

/* *** FLOATING POINT GARBAGE ********************************************* */

static void build_glulx_float(double f, unsigned int *w1)
{
    /* FIXME */
    *w1 = 0;
}

static void build_glulx_double(double f, unsigned int *w1, unsigned int *w2)
{
    /* FIXME */    
    *w1 = 0;
    *w2 = 0;
}
